import os
import re
import dis
import sys
import json
import time
import socket
import threading
import configparser

from select import select
from datetime import datetime
from PyQt5.QtWidgets import QApplication

from server_app.common.variables import (MAX_CONNECTIONS, RESPONSE, ERROR, TIME, USER, ACTION, ACCOUNT_NAME, PRESENCE,
                                         MESSAGE, EXIT, TO_USERNAME, USERNAME_SERVER, USERS_ONLINE, ACTION_GET_CONTACTS,
                                         ACTION_ADD_CONTACT, ACTION_DEL_CONTACT, ACTION_GET_CLIENTS)
from server_app.common.utils import get_message, send_message, parse_cmd_parameter, PortField, result_from_stdout
from server_app.common.exceptions import CodeException
from server_app.common.decorators import login_required
from server_app.logs.server_log_config import server_log
from server_app.db.server_storage import ServerStorage
from server_app.views.server_gui import ServerGui


class ServerVerifier(type):

    def __init__(self, *args, **kwargs):

        super(ServerVerifier, self).__init__(*args, **kwargs)

        re_tcp = r'.*LOAD_ATTR.*SOCK_STREAM.*'
        re_connect = r'.*LOAD_METHOD.*connect.*'

        result_string = result_from_stdout(dis.dis, self)

        if not re.search(re_tcp, result_string):
            raise CodeException('Допустимы только TCP сокеты.')

        if re.search(re_connect, result_string):
            raise CodeException('Вызовы метода connect недопустимы.')


class Server(metaclass=ServerVerifier):
    """
    Класс сервер
    """

    __listen_port = PortField()

    def __init__(self, listen_address, listen_port, config_file_path):
        self.__clients_online_db = {}
        self.__listen_address = listen_address
        self.__listen_port = listen_port
        self.__storage = ServerStorage()
        self.__config_file_path = config_file_path
        self.__clients_sockets = []

# region protocol

    def process_client_message(self, message):
        """
        Обработчик сообщений от клиентов, принимает словарь -
        сообщение от клиента, проверяет корректность,
        возвращает словарь-ответ для клиента
        """

        if ACTION in message and message[ACTION] == PRESENCE and TIME in message and USER in message:
            return {RESPONSE: 200, MESSAGE: message}

        if ACTION in message and message[ACTION] == MESSAGE and TIME in message and USER in message:
            return {RESPONSE: 201, MESSAGE: message}

        if ACTION in message and message[ACTION] == EXIT and TIME in message and USER in message:
            return {RESPONSE: 202, MESSAGE: message}

        if ACTION in message and message[ACTION] == USERS_ONLINE and TIME in message and USER in message:
            return {RESPONSE: 203, MESSAGE: message}

        if ACTION in message and message[ACTION] == ACTION_GET_CONTACTS and TIME in message and USER in message:
            return {RESPONSE: 204, MESSAGE: message}

        if ACTION in message and message[ACTION] == ACTION_ADD_CONTACT and TIME in message and USER in message:
            return {RESPONSE: 205, MESSAGE: message}

        if ACTION in message and message[ACTION] == ACTION_DEL_CONTACT and TIME in message and USER in message:
            return {RESPONSE: 206, MESSAGE: message}

        if ACTION in message and message[ACTION] == ACTION_GET_CLIENTS and TIME in message and USER in message:
            return {RESPONSE: 207, MESSAGE: message}

        return {
            RESPONSE: 400,
            ERROR: 'Bad Request'
        }

    def create_common_message(self, response, client):
        return {
            RESPONSE: response,
            TIME: time.time(),
            USER: client,
            MESSAGE: ''
        }

    def create_ok_answer(self):
        """
        Генерирует ответ на приветствие
        :return:
        """

        message = self.create_common_message(301, USERNAME_SERVER)
        message[MESSAGE] = 'OK'

        return message

    def create_no_client_answer(self):
        """
        Генерирует сообщение пользователь не найден
        :param response:
        :return:
        """

        message = self.create_common_message(201, USERNAME_SERVER)
        message[MESSAGE] = 'Пользователь не найден, возможно он не в сети, или вы ошиблись в имени.'

        return message

    def create_no_register_client_answer(self):
        """
        Генерирует сообщение пользователь не зарегистрирован
        :param response:
        :return:
        """

        message = self.create_common_message(208, USERNAME_SERVER)
        message[MESSAGE] = 'Пользователь не зарегистрирован.'

        return message

    def create_invalid_pwd_client_answer(self):
        """
        Генерирует сообщение пароль не верный
        :param response:
        :return:
        """

        message = self.create_common_message(209, USERNAME_SERVER)
        message[MESSAGE] = 'Пароль не верный.'

        return message

    def create_client_online_answer(self):
        """
        Генерирует сообщение пользователи онлайн
        :param response:
        :return:
        """

        message = self.create_common_message(203, USERNAME_SERVER)

        message_text = 'Список пользователей онлайн: \n'
        message_text += '\n'.join(['/' + client for client in self.__clients_online_db])

        message[MESSAGE] = message_text

        return message

    def create_get_clients_answer(self):
        clients = self.__storage.get_register_clients()
        message = self.create_common_message(207, USERNAME_SERVER)
        message[MESSAGE] = clients

        return message

    def create_client_contacts_answer(self, client_name):
        """
        Генерирует сообщение пользователи онлайн
        :param response:
        :return:
        """

        message = self.create_common_message(203, USERNAME_SERVER)

        client_id = self.__storage.get_client_id(client_name)
        contacts = self.__storage.get_contacts(client_id)

        message[MESSAGE] = contacts

        return message

    def create_answer(self, response):
        """
        Генерирует сообщение пользователю
        :param response:
        :return:
        """

        message = self.create_common_message(201, response[MESSAGE][USER][ACCOUNT_NAME])
        message[MESSAGE] = response[MESSAGE][MESSAGE]

        return message

    def create_exit_answer(self, response):
        """
        Генерирует сообщение выхода
        :param response:
        :return:
        """

        message = self.create_common_message(202, response[MESSAGE][USER][ACCOUNT_NAME])
        message[MESSAGE] = 'exit and say by!'

        return message

    def create_add_contact_answer(self):
        message = self.create_common_message(300, USERNAME_SERVER)
        message[MESSAGE] = 'Ok.'

        return message

    def create_del_contact_answer(self):
        message = self.create_common_message(300, USERNAME_SERVER)
        message[MESSAGE] = 'Ok.'

        return message

# endregion

# region db

    def get_clients_online(self):
        clients_online = self.__storage.get_clients_online()
        result = 'Список пользователей онлайн: \n'
        for elem in clients_online:
            result += f'login: {str(elem[2])}, ip: {elem[0]}, info: {elem[1]}' + '\n'
        return result

    def add_client_contact(self, client, contact):
        # TODO закешировать
        client_id = self.__storage.get_client_id(client)
        contact_id = self.__storage.get_client_id(contact)

        self.__storage.add_contact(client_id, contact_id)

        return True

    def del_client_contact(self, client, contact):
        # TODO закешировать
        client_id = self.__storage.get_client_id(client)
        contact_id = self.__storage.get_client_id(contact)

        self.__storage.del_contact(client_id, contact_id)

    def get_history(self, client):
        # TODO закешировать
        client_id = self.__storage.get_client_id(client)

        history = self.__storage.get_history(client_id)
        result = 'История: \n'
        for elem in history:
            result += str(elem) + '\n'
        return result

    def clear_online(self):
        self.__storage.clear_online()

    def register_client_online(self, client, socket, ip_address, port):
        # TODO закешировать
        client_id = self.__storage.get_client_id(client)

        self.__clients_online_db[client] = socket
        self.__storage.register_client_online(client_id, ip_address, port, datetime.now())

    def update_stat(self, sender, recipient):
        # TODO закешировать
        sender_id = self.__storage.get_client_id(sender)
        recipient_id = self.__storage.get_client_id(recipient)

        self.__storage.update_stat(sender_id, recipient_id)

    def register_client_action(self, client, action, info):
        # TODO закешировать
        client_id = self.__storage.get_client_id(client)

        self.__storage.register_client_action(client_id, action, info)

    def unregister_client_online(self, client):
        # TODO закешировать
        client_id = self.__storage.get_client_id(client)

        self.__storage.unregister_client_online(client_id)
        del self.__clients_online_db[client]

    def get_socket_on_clientname(self, to_client):
        return self.__clients_online_db.get(to_client.replace('/', ''))

# endregion

    def autorize(self, client_name, socket):

        client_message = get_message(socket)

        client = self.__storage.get_client(client_name)

        if not client:
            try:
                send_message(socket, self.create_no_register_client_answer())
            except OSError:
                pass

            time.sleep(1)
            self.__clients_sockets.remove(socket)
            socket.close()

        if client:
            pwd_hash = client_message[MESSAGE].encode('ascii')
            if not client.pwd_hash == pwd_hash:
                try:
                    send_message(socket, self.create_invalid_pwd_client_answer())
                except OSError:
                    pass

                time.sleep(1)
                self.__clients_sockets.remove(socket)
                socket.close()

        send_message(socket, self.create_ok_answer())

    @login_required
    def __process_clients_messages(self, transport, response, client_address, client_socket, cl_sock_read, cl_sock_write, message_pool):

        # Пока так, 200 это приветствие
        if response[RESPONSE] == 200 and client_socket in cl_sock_write:
            transport.settimeout(10)
            client_name = response[MESSAGE][USER][ACCOUNT_NAME]
            send_message(client_socket, response)
            self.autorize(client_name, client_socket)
            transport.settimeout(1)

            self.register_client_online(client_name, client_socket, client_address[0], client_address[1])

        # Пока так, 201 это сообщение
        if response[RESPONSE] == 201:
            client_socket = self.get_socket_on_clientname(response[MESSAGE][TO_USERNAME])

            if client_socket:
                message_pool.append((client_socket, self.create_answer(response)))
            else:
                message_pool.append((client_socket, self.create_no_client_answer()))

            sender_name = response[MESSAGE][USER][ACCOUNT_NAME]
            recipient_name = response[MESSAGE][TO_USERNAME].replace('/', '')

            self.update_stat(sender_name, recipient_name)
            self.register_client_action(sender_name, 'send message', recipient_name)

        # Пока так, 202 это выход
        if response[RESPONSE] == 202 and client_socket in cl_sock_write:
            self.__clients_sockets.remove(client_socket)
            client_socket.close()

            self.register_client_action(response[MESSAGE][USER][ACCOUNT_NAME], 'exit', str(client_address))
            self.unregister_client_online(response[MESSAGE][USER][ACCOUNT_NAME])

        # Пока так, 203 это запрос пользователей онлайн
        if response[RESPONSE] == 203 and client_socket in cl_sock_write:
            self.register_client_action(response[MESSAGE][USER][ACCOUNT_NAME], 'get online', str(client_address))
            message_pool.append((client_socket, self.create_client_online_answer()))

        # Пока так, 204 это запрос списка контактов
        if response[RESPONSE] == 204 and client_socket in cl_sock_write:
            client_name = response[MESSAGE][USER][ACCOUNT_NAME]
            self.register_client_action(client_name, 'get contacts', str(client_address))
            message_pool.append((client_socket, self.create_client_contacts_answer(client_name)))

        # Пока так, 205 это запрос на добавление в контакты
        if response[RESPONSE] == 205 and client_socket in cl_sock_write:
            client_name = response[MESSAGE][USER][ACCOUNT_NAME]
            contacts_name = response[MESSAGE][TO_USERNAME]

            self.add_client_contact(client_name, contacts_name)

            self.register_client_action(client_name, 'add contact', str(client_address))
            message_pool.append((client_socket, self.create_add_contact_answer()))

        # Пока так, 206 это запрос на удаление из контактов
        if response[RESPONSE] == 206 and client_socket in cl_sock_write:
            client_name = response[MESSAGE][USER][ACCOUNT_NAME]
            contacts_name = response[MESSAGE][TO_USERNAME]

            self.del_client_contact(client_name, contacts_name)

            self.register_client_action(client_name, 'del contact', str(client_address))
            message_pool.append((client_socket, self.create_del_contact_answer()))

        # Пока так, 207 это запрос на список клиентов
        if response[RESPONSE] == 207 and client_socket in cl_sock_write:
            client_name = response[MESSAGE][USER][ACCOUNT_NAME]
            self.register_client_action(client_name, 'get clients', str(client_address))
            message_pool.append((client_socket, self.create_get_clients_answer()))

    def __process_messages(self):
        """
        Для потока обработки сообщений
        :return:
        """

        transport = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        transport.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        transport.bind((self.__listen_address, self.__listen_port))

        transport.listen(MAX_CONNECTIONS)
        transport.settimeout(1)

        self.clear_online()

        self.__clients_sockets = []
        server_log.info(f'Сервер запущен по адресу: {self.__listen_address}: {self.__listen_port}')

        while True:
            try:
                client_sock, client_address = transport.accept()
            except OSError as e:
                pass
            else:
                self.__clients_sockets.append(client_sock)
            finally:
                wait = 0
                message_pool = []

                for client_socket in self.__clients_sockets:
                    # На мой взгляд логичнее это вынести за цикл, но на виндовс так не работает
                    cl_sock_read, cl_sock_write = [], []
                    cl_sock_read, cl_sock_write, _ = select(self.__clients_sockets, self.__clients_sockets, [], wait)
                    try:
                        if client_socket in cl_sock_read:
                            client_message = get_message(client_socket)
                            response = self.process_client_message(client_message)

                            self.__process_clients_messages(transport, response, client_address, client_socket, cl_sock_read, cl_sock_write, message_pool)

                    except ConnectionResetError as e:
                        server_log.exception(f'Произошла ошибка: {str(e)}')
                        self.__clients_sockets.remove(client_socket)
                        client_socket.close()

                    except (ValueError, json.JSONDecodeError, TypeError):
                        server_log.exception('Принято некорректное сообщение от клиента')
                        self.__clients_sockets.remove(client_socket)
                        client_socket.close()

                for message in message_pool:
                    _, cl_sock_write, _ = select([], self.__clients_sockets, [], wait)

                    client_socket = message[0]
                    message_send = message[1]

                    if client_socket in cl_sock_write:
                        send_message(client_socket, message_send)

    def __process_gui(self):
        """
        Для потока администрирования
        :return:
        """

        server_app = QApplication(sys.argv)

        server_gui = ServerGui(self.__config_file_path)
        server_gui.set_timer(3000)
        server_gui.show()

        server_gui.status_message('Welcome, admin. SHODAN is waiting you.')

        server_app.exec()

    def run(self):
        """
        Запускает сервер.
        Пример: server.py -p 8888 -a 127.0.0.1
        """

        process_messages = threading.Thread(target=self.__process_messages, daemon=True)
        process_gui = threading.Thread(target=self.__process_gui, daemon=True)

        process_messages.daemon = True
        process_gui.daemon = True

        process_messages.start()
        process_gui.start()

        while True:
            time.sleep(1)
            if process_messages.is_alive() and process_gui.is_alive():
                continue
            break


if __name__ == '__main__':

    config = configparser.ConfigParser()
    dir_path = os.path.dirname(os.path.relpath(__file__))
    config_file_path = os.path.join(dir_path, 'settings.ini')
    config.read(config_file_path)

    listen_address = parse_cmd_parameter('-a', sys.argv, config['SETTINGS']['default_address'],
                                         'После параметра \'a\'- необходимо указать адрес, который будет слушать сервер.')
    listen_port = parse_cmd_parameter('-p', sys.argv, config['SETTINGS']['default_port'],
                                      'После параметра -\'p\' необходимо указать номер порта.')

    listen_port = int(listen_port)

    if listen_port is None or listen_address is None:
        raise ValueError('Неверно заданы параметры командной строки')

    server = Server(listen_address, int(listen_port), config_file_path)
    server.run()



import os
import configparser

from PyQt5 import uic
from PyQt5.QtCore import QTimer
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtWidgets import QMainWindow, QLineEdit, QMessageBox
from datetime import datetime

from server_app.db.server_storage import ServerStorage
from server_app.common.utils import generate_hash


FORM_CLASS, _ = uic.loadUiType(os.path.join(os.path.dirname(__file__), 'server_gui.ui'))


class ServerGui(QMainWindow, FORM_CLASS):
    
    def __init__(self, config_file_path):
        super(ServerGui, self).__init__()

        self.setupUi(self)

        self.__table_clients_online_model = QStandardItemModel()
        self.__table_clients_stat_model = QStandardItemModel()
        self.__table_clients_online_header = ['Клиент', 'IP адрес', 'Порт', 'Подключен']
        self.__table_clients_stat_header = ['Клиент', 'Вход', 'Отправлено', 'Получено']

        self.__timer = None
        self.__storage = ServerStorage()
        self.__config_file_path = config_file_path
        self.__messages = QMessageBox()

        self.__load_config()
        self.__load_client_list()

        self.initUi()

    def initUi(self):
        self.__table_clients_online_model.setHorizontalHeaderLabels(self.__table_clients_online_header)
        self.__table_clients_stat_model.setHorizontalHeaderLabels(self.__table_clients_stat_header)

        self.btn_save.clicked.connect(self.__save_config)
        self.table_clients_online.setModel(self.__table_clients_online_model)
        self.table_clients_stat.setModel(self.__table_clients_stat_model)

        self.lineEdit_pwd1.setEchoMode(QLineEdit.Password)
        self.lineEdit_pwd2.setEchoMode(QLineEdit.Password)

        self.pushButton_add_client.clicked.connect(self.__add_client_clicked)
        self.pushButton_del_client.clicked.connect(self.__del_client_clicked)

    def __load_config(self):
        config = configparser.ConfigParser()
        config.read(self.__config_file_path)

        self.edt_database_dir.setText(config['SETTINGS']['database_dir'])
        self.edt_database_file.setText(config['SETTINGS']['database_file'])
        self.edt_default_address.setText(config['SETTINGS']['default_address'])
        self.edt_default_port.setText(config['SETTINGS']['default_port'])

    def __save_config(self):
        config = configparser.ConfigParser()
        config.read(self.__config_file_path)

        config['SETTINGS']['database_dir'] = self.edt_database_dir.text()
        config['SETTINGS']['database_file'] = self.edt_database_file.text()
        config['SETTINGS']['default_address'] = self.edt_default_address.text()
        config['SETTINGS']['default_port'] = self.edt_default_port.text()

        with open(self.__config_file_path, 'w') as f:
            config.write(f)

        self.status_message('Сохранено')

    def __on_timer_tic(self):
        self.load_client_table()
        self.load_stat_table()

    def __add_client_clicked(self):
        client_name = self.lineEdit_client_name.text().strip()
        pwd1 = self.lineEdit_pwd1.text()
        pwd2 = self.lineEdit_pwd2.text()

        if not client_name:
            self.status_message('Имя пользователя не указано.')
            return

        if not pwd1 == pwd2:
            self.status_message('Пароли не совпадают.')
            return

        if self.__storage.get_client_id(client_name):
            self.status_message('Такой пользователь уже существует.')
            return

        self.__storage.add_client(client_name, generate_hash(client_name, pwd1), datetime.now())

        self.lineEdit_client_name.setText('')
        self.lineEdit_pwd1.setText('')
        self.lineEdit_pwd2.setText('')

        self.__load_client_list()

        self.status_message('Пользователь добавлен.')

    def __del_client_clicked(self):
        if self.__messages.question(self, 'Удаление контакта', 'Вы уверены?', QMessageBox.Yes,
                                  QMessageBox.No) == QMessageBox.No:
            return

        client_name = self.comboBox_clients.currentText()
        client_id = self.__storage.get_client_id(client_name)

        if not client_name:
            self.status_message('Пользователь не найден, возможно он уже удален.')
            self.__load_client_list()
            return

        self.__storage.del_client(client_id)
        self.__load_client_list()

    def __load_client_list(self):
        self.comboBox_clients.clear()
        self.comboBox_clients.addItems(self.__storage.get_register_clients())

    def status_message(self, message):
        self.statusbar.showMessage(message)

    def load_client_table(self):
        self.__table_clients_online_model.clear()
        self.__table_clients_online_model.setHorizontalHeaderLabels(self.__table_clients_online_header)

        clients_online = self.__storage.get_clients_online()
        for elem in clients_online:
            ip, port, info, client = elem

            client = QStandardItem(client)
            ip = QStandardItem(ip)
            port = QStandardItem(port)
            info = QStandardItem(info)

            client.setEditable(False)
            ip.setEditable(False)
            port.setEditable(False)
            info.setEditable(False)

            self.__table_clients_online_model.appendRow([client, ip, port, info])

    def load_stat_table(self):
        self.__table_clients_stat_model.clear()
        self.__table_clients_stat_model.setHorizontalHeaderLabels(self.__table_clients_stat_header)

        stats = self.__storage.get_stat()
        for elem in stats:
            client, info, sent, recv = elem

            client = QStandardItem(client)
            info = QStandardItem(info)
            sent = QStandardItem(str(sent))
            recv = QStandardItem(str(recv))

            client.setEditable(False)
            info.setEditable(False)
            sent.setEditable(False)
            recv.setEditable(False)

            self.__table_clients_stat_model.appendRow([client, info, sent, recv])

    def set_timer(self, interval):
        self.__timer = QTimer()
        self.__timer.timeout.connect(self.__on_timer_tic)
        self.__timer.start(interval)



from sqlalchemy import Column, Integer, String, ForeignKey, DateTime
from server_app.db.db_connect import Base


class DbHistory(Base):
    __tablename__ = 'history'
    __table_args__ = {'extend_existing': True}

    id = Column(Integer, primary_key=True)
    client_id = Column(Integer, ForeignKey('clients.id', ondelete='CASCADE'))
    date_action = Column(DateTime)
    action = Column(String)
    info = Column(String)

    def __init__(self, client_id, date_action, action, info):
        self.client_id = client_id
        self.date_action = date_action
        self.action = action
        self.info = info

    def __repr__(self):
        return f'< {self.client_id}, {self.date_action}, {self.action}, {self.info}>'


from sqlalchemy import Column, Integer, ForeignKey
from server_app.db.db_connect import Base


class DbStat(Base):
    __tablename__ = 'stat'
    __table_args__ = {'extend_existing': True}

    id = Column(Integer, primary_key=True)
    client_id = Column(Integer, ForeignKey('clients.id', ondelete='CASCADE'))
    sent = Column(Integer)
    recv = Column(Integer)

    def __init__(self, client_id):
        self.client_id = client_id

    def __repr__(self):
        return f'< {self.client_id}, {self.sent}, {self.recv}>'


from sqlalchemy import create_engine
from sqlalchemy.orm import declarative_base, sessionmaker


engine = create_engine('sqlite:///server.db3', connect_args={'check_same_thread': False})
# engine = create_engine('sqlite:///server.db3', echo=True, connect_args={'check_same_thread': False})
# engine = create_engine('sqlite:///server.db3', echo=True)

Base = declarative_base()

Session = sessionmaker(bind=engine)

session = Session()



from sqlalchemy import Column, Integer, String, ForeignKey
from server_app.db.db_connect import Base


class DbClientsOnline(Base):
    __tablename__ = 'clients_online'
    __table_args__ = {'extend_existing': True}

    id = Column(Integer, primary_key=True)
    client_id = Column(Integer, ForeignKey('clients.id', ondelete='CASCADE'))
    ip_address = Column(String)
    port = Column(String)
    info = Column(String)

    def __init__(self, client_id, ip_address, port, info):
        self.client_id = client_id
        self.ip_address = ip_address
        self.port = port
        self.info = info

    def __repr__(self):
        return f'client_id: {self.client_id}, info: { self.info}'


from server_log_config import server_log

import traceback


class Log():
    """ Декоратор лога классом """

    def __init__(self):
        self.__loggi = server_log

    def __call__(self, func):
        def wrapper(*args, **kwargs):
            from_func_name = traceback.format_stack()[0].strip().split()[-1]
            self.__loggi.debug(f'Функция "{func.__name__}" вызвана c параметрами "args: {args}", "kwargs: {kwargs}"')
            self.__loggi.debug(f'Функция "{func.__name__}" вызвана из "{from_func_name}"')

            result = func(*args, **kwargs)

            return result

        return wrapper


def log(func):
    """ Декоратор лога функцией """

    loggi = server_log

    def wrapper(*args, **kwargs):
        from_func_name = traceback.format_stack()[0].strip().split()[-1]
        loggi.debug(f'Функция "{func.__name__}" вызвана c параметрами "args: {args}", "kwargs: {kwargs}"')
        loggi.debug(f'Функция "{func.__name__}" вызвана из "{from_func_name}"')

        result = func(*args, **kwargs)

        return result

    return wrapper


from sqlalchemy import Column, Integer, String
from server_app.db.db_connect import Base


class DbClient(Base):
    __tablename__ = 'clients'
    __table_args__ = {'extend_existing': True}

    id = Column(Integer, primary_key=True)
    login = Column(String)
    pwd_hash = Column(String)
    info = Column(String)

    def __init__(self, login, pwd_hash, info):
        self.login = login
        self.info = info
        self.pwd_hash = pwd_hash

    def __repr__(self):
        return f'id: {self.id}, login: {self.login}, last login: {self.info}'


from sqlalchemy import Column, Integer, String, ForeignKey
from server_app.db.db_connect import Base


class DbContacts(Base):
    __tablename__ = 'contacts'
    __table_args__ = {'extend_existing': True}

    id = Column(Integer, primary_key=True)
    client_id = Column(Integer, ForeignKey('clients.id', ondelete='CASCADE'))
    contact_id = Column(Integer, ForeignKey('clients.id', ondelete='CASCADE'))
    info = Column(String)

    def __init__(self, client_id, contact_id, info=''):
        self.client_id = client_id
        self.contact_id = contact_id
        self.info = info

    def __repr__(self):
        return f'id: {self.id}, client_id: {self.client_id}, contact_id: {self.contact_id}'


""" Константы, общие значения для всего прилоежения """

STRANGE_STRING = '4c4e86fd-5a13-4c20-93ff-c62ef5153d7c'

# Максимальная очередь подключений
MAX_CONNECTIONS = 5
# Максимальная длинна сообщения в байтах
MAX_PACKAGE_LENGTH = 4096
# Кодировка проекта
ENCODING = 'utf-8'

# Протокол JIM основные ключи:
ACTION = 'action'
TIME = 'time'
USER = 'user'
ACCOUNT_NAME = 'account_name'

ACTION_GET_CONTACTS = 'get_contacts'
ACTION_GET_CLIENTS = 'get_clients'
ACTION_ADD_CONTACT = 'add_contact'
ACTION_DEL_CONTACT = 'del_contact'

# Прочие ключи, используемые в протоколе
USERNAME_SERVER = 'server'
USERS_ONLINE = 'users_online'
TO_USERNAME = 'to_username'
PRESENCE = 'presence'
MESSAGE = 'message'
RESPONSE = 'response'
RESPONSE_OK = 'response_ok'
ERROR = 'error'
EXIT = 'exit'


""" Общие процедуры и фукнции """

import sys
import json
import hashlib
import binascii

from io import StringIO

from server_app.common.variables import MAX_PACKAGE_LENGTH, ENCODING, STRANGE_STRING
from server_app.logs.decorators import log, Log


class PortField:
    """
    Дескриптор для портов
    """

    def __get__(self, instance, owner):
        return instance.__dict__[self.my_attr]

    def __set__(self, instance, value):
        if value < 1024 or value > 65535:
             raise ValueError(f'{self.my_attr} - может быть только в диапазоне от 1024 до 65535.')
        instance.__dict__[self.my_attr] = value

    def __delete__(self, instance):
        del instance.__dict__[self.my_attr]

    def __set_name__(self, owner, my_attr):
        self.my_attr = my_attr


def generate_hash(client_name, password):
    """
    Генерирует хэш
    """

    passwd_bytes = password.encode('utf-8')
    salt = (client_name + STRANGE_STRING).encode('utf-8')
    password_hash = hashlib.pbkdf2_hmac('sha512', passwd_bytes, salt, 10000)

    return binascii.hexlify(password_hash)


def result_from_stdout(func, arg):
    """
    Перенаправляет поток вывода функции в переменную
    """

    old_stdout = sys.stdout
    result = StringIO()
    sys.stdout = result
    func(arg)
    result_string = result.getvalue()
    sys.stdout = old_stdout

    return result_string


@Log()
def parse_cmd_parameter(parameter, sys_argv, default_value, error_message):
    """
    Извлекает параметр из командной строки, если он не задан, возвращает default_value, в случае неудачи
    возвращает error_message
    """

    try:
        if not isinstance(sys_argv, list):
            raise TypeError
        if parameter in sys_argv:
            result = sys_argv[sys_argv.index(parameter) + 1]
        else:
            result = default_value

    except TypeError:
        result = None
        print('Второй параметр должен быть списком')
    except IndexError:
        result = None
        print(error_message)

    return result


@log
def send_message(socket, message):
    """
    Отправляет сообщение через сокет
    """

    if not isinstance(message, dict):
        raise TypeError

    json_string = json.dumps(message)
    message_bytes = json_string.encode(ENCODING)
    socket.send(message_bytes)


@log
def get_message(socket):
    """
    Получает сообщение из сокета, возвращает словарь с информацией о сообщении
    в случае ошибки выбрасывает ValueError
    """

    message_bytes = socket.recv(MAX_PACKAGE_LENGTH)
    if isinstance(message_bytes, bytes):
        json_string = message_bytes.decode(ENCODING)
        if isinstance(json_string, str):
            if not json_string:
                return {}
            message = json.loads(json_string)
            if isinstance(message, dict):
                return message

    raise ValueError



import os
import logging

from logging.handlers import TimedRotatingFileHandler

server_log = logging.getLogger('server.main')
formatter = logging.Formatter("<%(asctime)s> <%(levelname)s> <%(module)s> <%(message)s>")

file_name = os.path.dirname(os.path.abspath(__file__))
file_name = os.path.join(file_name, 'server.log')

rotation_file_handler = TimedRotatingFileHandler(file_name, when='d', interval=1, backupCount=6, encoding='utf-8')
rotation_file_handler.setLevel(logging.DEBUG)
rotation_file_handler.setFormatter(formatter)

if not server_log.handlers:
    server_log.addHandler(rotation_file_handler)
server_log.setLevel(logging.DEBUG)


from setuptools import setup, find_packages

setup(name="monster-chat-server",
      version="1.0.1",
      description="MMMMonsterChat server application",
      author="KTo ETo",
      author_email="kto@lovetou.ru",
      packages=find_packages(),
      install_requires=['PyQt5', 'sqlalchemy', 'pycryptodome', 'pycryptodomex'],
      scripts=['server_app/mc_server_start']
      )



from sqlalchemy import update, select, and_
from datetime import datetime

from server_app.db.db_connect import session, engine
from server_app.db.db_client import DbClient
from server_app.db.db_history import DbHistory
from server_app.db.db_clients_online import DbClientsOnline
from server_app.db.db_contacts import DbContacts
from server_app.db.db_stat import DbStat


DbClient.metadata.create_all(engine)
DbHistory.metadata.create_all(engine)
DbClientsOnline.metadata.create_all(engine)
DbContacts.metadata.create_all(engine)
DbStat.metadata.create_all(engine)


class ServerStorage:

    def get_client_id(self, client_login):
        data = session.query(DbClient).filter(DbClient.login == client_login).limit(1).first()
        return data.id if data else 0

    def get_client(self, client_login):
        data = session.query(DbClient).filter(DbClient.login == client_login).limit(1).first()
        return data

    def add_client(self, client, pwd_hash, info=''):
        client = DbClient(client, pwd_hash, info)
        session.add(client)
        session.commit()

        return True

    def update_client(self, client_id, info=''):
        upd = update(DbClient)
        upd = upd.values({'info': info})
        upd = upd.where(DbClient.id == client_id)
        engine.execute(upd)

    def del_client(self, client_id):
        session.query(DbClient).filter(DbClient.id == client_id).delete()
        session.commit()

    def get_register_clients(self):
        data = []
        stm = session.query(DbClient.login).all()
        for row in stm:
            data.append(row.login)
        return data

    def register_client_online(self, client_id, ip_address, port, info):
        client_online = DbClientsOnline(client_id, ip_address, port, info)
        session.add(client_online)
        session.commit()

    def unregister_client_online(self, client_id):
        session.query(DbClientsOnline).filter(DbClientsOnline.client_id == client_id).delete()
        session.commit()

    def register_client_action(self, client_id, action, info):
        history = DbHistory(client_id, datetime.now(), action, info)
        session.add(history)
        session.commit()

    def get_clients_online(self):
        stm = select(DbClientsOnline.ip_address, DbClientsOnline.port, DbClientsOnline.info, DbClient.login).join(DbClient,
                                                                                            DbClientsOnline.client_id == DbClient.id,
                                                                                            isouter=True)
        data = []
        result = session.execute(stm)

        for row in result:
            data.append(row)

        return data

    def get_history(self, client_id):
        data = []
        if client_id:
            stm = select(DbClient.login).where(DbHistory.client_id == client_id).join(DbClient,
                                                                                      DbHistory.client_id == DbClient.id,
                                                                                      isouter=True)
        else:
            stm = select(DbClient.login).join(DbClient, DbHistory.client_id == DbClient.id, isouter=True)

        result = session.execute(stm)

        for row in result:
            data.append(row)

        return data

    def update_stat(self, sender_id, recipient_id):
        sender_row = session.query(DbStat).filter(DbStat.client_id == sender_id).limit(1).first()
        recipient_row = session.query(DbStat).filter(DbStat.client_id == recipient_id).limit(1).first()

        if sender_row:
            sender_row.sent += 1
        else:
            sender = DbStat(sender_id)
            sender.sent, sender.recv = 1, 0
            session.add(sender)

        if recipient_row:
            recipient_row.recv += 1
        else:
            recipient = DbStat(recipient_id)
            recipient.sent, recipient.recv = 0, 1
            session.add(recipient)

        session.commit()

    def get_stat(self):
        data = []
        stm = select(DbClient.login, DbClient.info, DbStat.sent, DbStat.recv).join(DbClient, DbStat.client_id == DbClient.id, isouter=True)

        result = session.execute(stm)

        for row in result:
            data.append(row)

        return data

    def clear_online(self):
        session.query(DbClientsOnline).delete()
        session.commit()

    def add_contact(self, client_id, contact_id):
        exist = session.query(DbContacts).filter(
            and_(DbContacts.client_id == client_id, DbContacts.contact_id == contact_id)).limit(1).first()
        if not exist:
            contact = DbContacts(client_id, contact_id)
            session.add(contact)
            session.commit()

    def del_contact(self, client_id, contact_id):
        exist = session.query(DbContacts).filter(DbContacts.client_id == client_id).limit(1).first()
        if exist:
            session.query(DbContacts).where(
                and_(DbContacts.client_id == client_id, DbContacts.contact_id == contact_id)).delete()
            session.commit()

    def get_contacts(self, client_id):
        data = []
        stm = select(DbClient.login, DbContacts).where(DbContacts.client_id == client_id).join(DbClient,
                                                                                               DbContacts.contact_id == DbClient.id,
                                                                                               isouter=True)

        result = session.execute(stm)

        for row in result:
            data.append(row.login)

        return data



""" Декораторы сервера"""

def login_required(func):
    """
    Декоратор, проверяющий, что клиент авторизован на сервере.
    Проверяет, что передаваемый объект сокета находится в
    списке авторизованных клиентов.
    За исключением передачи словаря-запроса
    на авторизацию. Если клиент не авторизован,
    генерирует исключение TypeError
    """

    def checker(*args, **kwargs):
        if func.__name__ == '__process_clients_messages':
            from server_app.common.variables import MESSAGE, ACTION, USER, ACCOUNT_NAME
            register_clients = args[0]._Server__clients_online_db
            client_name = args[2][MESSAGE][USER][ACCOUNT_NAME]
            action = args[2][MESSAGE][ACTION]

            if not action == 'presence':
                if not register_clients.get(client_name):
                    raise TypeError

        return func(*args, **kwargs)

    return checker
